@@ -36,6 +36,7 @@ from distutils.util import strtobool
 from fnmatch import fnmatch
 from importlib.abc import Loader
 from typing import Any, DefaultDict, Dict, Iterator, List, Set, Tuple
+from unittest.mock import MagicMock, patch
 
 import boto3
 import botocore
@@ -44,25 +45,14 @@ import semver
 from ruamel.yaml import YAML
 from ruamel.yaml import parser as YAMLParser
 from ruamel.yaml import scanner as YAMLScanner
-from schema import (
-    Optional,
-    Schema,
-    SchemaError,
-    SchemaForbiddenKeyError,
-    SchemaMissingKeyError,
-    SchemaUnexpectedTypeError,
-    SchemaWrongKeyError,
-)
+from schema import (Optional, Schema, SchemaError, SchemaForbiddenKeyError,
+                    SchemaMissingKeyError, SchemaUnexpectedTypeError,
+                    SchemaWrongKeyError)
 
-from panther_analysis_tool.schemas import (
-    DATA_MODEL_SCHEMA,
-    GLOBAL_SCHEMA,
-    PACK_SCHEMA,
-    POLICY_SCHEMA,
-    RULE_SCHEMA,
-    SCHEDULED_QUERY_SCHEMA,
-    TYPE_SCHEMA,
-)
+from panther_analysis_tool.schemas import (DATA_MODEL_SCHEMA, GLOBAL_SCHEMA,
+                                           PACK_SCHEMA, POLICY_SCHEMA,
+                                           RULE_SCHEMA, SCHEDULED_QUERY_SCHEMA,
+                                           TYPE_SCHEMA)
 from panther_analysis_tool.test_case import DataModel, TestCase
 
 DATA_MODEL_LOCATION = "./data_models"
@@ -84,6 +74,14 @@ QUERY = "scheduled_query"
 SCHEDULED_RULE = "scheduled_rule"
 RULE = "rule"
 
+RESERVED_FUNCTIONS = [
+    "dedup", "title", "description", "reference", "severity", "runbook", "destinations"
+]
+
+VALID_SEVERITIES = [
+    "INFO", "LOW", "MEDIUM", "HIGH", "CRITICAL"
+]
+
 SCHEMAS: Dict[str, Schema] = {
     DATAMODEL: DATA_MODEL_SCHEMA,
     GLOBAL: GLOBAL_SCHEMA,
@@ -94,6 +92,15 @@ SCHEMAS: Dict[str, Schema] = {
     SCHEDULED_RULE: RULE_SCHEMA,
 }
 
+SET_FIELDS = [
+    "LogTypes",
+    "PackIDs",
+    "OutputIds",
+    "SummaryAttributes",
+    "Suppressions",
+    "Tags",
+]
+
 
 # exception for conflicting ids
 class AnalysisIDConflictException(Exception):
@@ -102,6 +109,14 @@ class AnalysisIDConflictException(Exception):
         super().__init__(self.message)
 
 
+# exception for conflicting ids
+class AnalysisContainsDuplicatesException(Exception):
+    def __init__(self, analysis_id: str, invalid_fields: List[str]):
+        self.message = "Specification file for [{}] contains fields with duplicate values: [{}]" \
+            .format(analysis_id, ', '.join(x for x in invalid_fields))
+        super().__init__(self.message)
+
+
 def load_module(filename: str) -> Tuple[Any, Any]:
     """Loads the analysis function module from a file.
 
@@ -141,6 +156,10 @@ def load_analysis_specs(directories: List[str]) -> Iterator[Tuple[str, str, Any,
     loaded_specs: List[Any] = []
     for directory in directories:
         for relative_path, _, file_list in os.walk(directory):
+            # Skip hidden folders
+            if relative_path.split('/')[-1].startswith(".") and \
+                relative_path != "./" and relative_path != ".":
+                continue
             # setup yaml object
             yaml = YAML(typ="safe")
             # If the user runs with no path args, filter to make sure
@@ -157,13 +176,17 @@ def load_analysis_specs(directories: List[str]) -> Iterator[Tuple[str, str, Any,
                             RULES_PATH_PATTERN,
                             PACKS_PATH_PATTERN,
                             POLICIES_PATH_PATTERN,
+                            QUERIES_PATH_PATTERN,
                         )
                     )
                 ):
                     logging.debug("Skipping path %s", relative_path)
                     continue
             for filename in sorted(file_list):
-                spec_filename = os.path.join(relative_path, filename)
+                # Skip hidden files
+                if filename.startswith("."):
+                    continue
+                spec_filename = os.path.abspath(os.path.join(relative_path, filename))
                 # skip loading files that have already been imported
                 if spec_filename in loaded_specs:
                     continue
@@ -573,12 +596,20 @@ def test_analysis(args: argparse.Namespace) -> Tuple[int, list]:
     Returns:
         A tuple of the return code, and a list of tuples containing invalid specs and their error.
     """
-    failed_tests: DefaultDict[str, list] = defaultdict(list)
     logging.info("Testing analysis packs in %s\n", args.path)
 
+    search_directories = [args.path]
+
+    # Try the parent directory as well
+    for directory in (HELPERS_LOCATION, '.' + HELPERS_LOCATION,
+                DATA_MODEL_LOCATION, '.' + DATA_MODEL_LOCATION):
+        absolute_dir_path = os.path.abspath(os.path.join(args.path, directory))
+        if os.path.exists(absolute_dir_path):
+            search_directories.append(absolute_dir_path)
+
     # First classify each file, always include globals and data models location
     specs, invalid_specs = classify_analysis(
-        list(load_analysis_specs([args.path, HELPERS_LOCATION, DATA_MODEL_LOCATION]))
+        list(load_analysis_specs(search_directories))
     )
 
     if all((len(specs[key]) == 0 for key in specs)):
@@ -678,7 +709,7 @@ def setup_run_tests(
             invalid_specs.append((analysis_spec_filename, load_err))
             continue
 
-        analysis_funcs = {}
+        analysis_funcs = {"module": module}
         if analysis_type == POLICY:
             analysis_funcs["run"] = module.policy
         elif analysis_type in [RULE, SCHEDULED_RULE]:
@@ -687,6 +718,16 @@ def setup_run_tests(
                 analysis_funcs["dedup"] = module.dedup
             if "title" in dir(module):
                 analysis_funcs["title"] = module.title
+            if "description" in dir(module):
+                analysis_funcs["description"] = module.description
+            if "reference" in dir(module):
+                analysis_funcs["reference"] = module.reference
+            if "severity" in dir(module):
+                analysis_funcs["severity"] = module.severity
+            if "runbook" in dir(module):
+                analysis_funcs["runbook"] = module.runbook
+            if "destinations" in dir(module):
+                analysis_funcs["destinations"] = module.destinations
 
         failed_tests = run_tests(
             analysis_spec,
@@ -753,6 +794,7 @@ def filter_analysis(analysis: List[Any], filters: Dict[str, List]) -> List[Any]:
     return filtered_analysis
 
 
+# pylint: disable=too-many-locals,too-many-statements
 def classify_analysis(
     specs: List[Tuple[str, str, Any, Any]]
 ) -> Tuple[Dict[str, List[Any]], List[Any]]:
@@ -769,8 +811,11 @@ def classify_analysis(
     # add any duplicates to the invalid_specs
     analysis_ids: List[Any] = []
 
+    # pylint: disable=too-many-nested-blocks
     for analysis_spec_filename, dir_name, analysis_spec, error in specs:
         keys: List[Any] = list()
+        tmp_logtypes: Any = None
+        tmp_logtypes_key: Any = None
         try:
             # check for parsing errors from json.loads (ValueError) / yaml.safe_load (YAMLError)
             if error:
@@ -781,11 +826,24 @@ def classify_analysis(
             # validate the particular analysis type schema
             analysis_schema = SCHEMAS[analysis_type]
             keys = list(analysis_schema.schema.keys())
+            # Special case for ScheduledQueries to only validate the types
+            if "ScheduledQueries" in analysis_spec:
+                for each_key in analysis_schema.schema.keys():
+                    if str(each_key) == "Or('LogTypes', 'ScheduledQueries')":
+                        tmp_logtypes_key = each_key
+                        break
+                if not tmp_logtypes:
+                    tmp_logtypes = analysis_schema.schema[tmp_logtypes_key]
+                analysis_schema.schema[tmp_logtypes_key] = [str]
             analysis_schema.validate(analysis_spec)
             # lookup the analysis type id and validate there aren't any conflicts
             analysis_id = lookup_analysis_id(analysis_spec, analysis_type)
             if analysis_id in analysis_ids:
                 raise AnalysisIDConflictException(analysis_id)
+            # check for duplicates where panther expects a unique set
+            invalid_fields = contains_invalid_field_set(analysis_spec)
+            if invalid_fields:
+                raise AnalysisContainsDuplicatesException(analysis_id, invalid_fields)
             analysis_ids.append(analysis_id)
             # add the validated analysis type to the classified specs
             if analysis_type in [POLICY, RULE, SCHEDULED_RULE]:
@@ -799,19 +857,33 @@ def classify_analysis(
         except SchemaWrongKeyError as err:
             invalid_specs.append((analysis_spec_filename, handle_wrong_key_error(err, keys)))
         except (
-            SchemaError,
             SchemaMissingKeyError,
             SchemaForbiddenKeyError,
             SchemaUnexpectedTypeError,
         ) as err:
             invalid_specs.append((analysis_spec_filename, err))
             continue
+        except SchemaError as err:
+            # Intercept the error, otherwise the error message becomes confusing and unreadable
+            error = err
+            err_str = str(err)
+            first_half = err_str.split(':')[0]
+            second_half = err_str.split(')')[-1]
+            if "LogTypes" in str(err):
+                error = SchemaError(f"{first_half}: LOG_TYPE_REGEX{second_half}")
+            elif "ResourceTypes" in str(err):
+                error = SchemaError(f"{first_half}: RESOURCE_TYPE_REGEX{second_half}")
+            invalid_specs.append((analysis_spec_filename, error))
         except Exception as err:  # pylint: disable=broad-except
             # Catch arbitrary exceptions thrown by bad specification files
             invalid_specs.append((analysis_spec_filename, err))
             continue
+        finally:
+            # Restore original values
+            if tmp_logtypes and tmp_logtypes_key:
+                analysis_schema.schema[tmp_logtypes_key] = tmp_logtypes
 
-    return (classified_specs, invalid_specs)
+    return classified_specs, invalid_specs
 
 
 def lookup_analysis_id(analysis_spec: Any, analysis_type: str) -> str:
@@ -831,6 +903,25 @@ def lookup_analysis_id(analysis_spec: Any, analysis_type: str) -> str:
     return analysis_id
 
 
+def contains_invalid_field_set(analysis_spec: Any) -> List[str]:
+    """Checks if the fields that Panther expects as sets have duplicates, returns True if invalid.
+
+    :param analysis_spec: Loaded YAML specification file
+    :return: bool - whether or not the specifications file is valid where False denotes valid.
+    """
+    invalid_fields = []
+    for field in SET_FIELDS:
+        if field not in analysis_spec:
+            continue
+        # Handle special case where we need to test for lowercase tags
+        if field == "Tags":
+            if len(analysis_spec[field]) != len(set(x.lower() for x in analysis_spec[field])):
+                invalid_fields.append("LowerTags")
+        if len(analysis_spec[field]) != len(set(analysis_spec[field])):
+            invalid_fields.append(field)
+    return invalid_fields
+
+
 def handle_wrong_key_error(err: SchemaWrongKeyError, keys: list) -> Exception:
     regex = r"Wrong key(?:s)? (.+?) in (.*)$"
     matches = re.match(regex, str(err))
@@ -864,52 +955,7 @@ def run_tests(
         print("\tNo tests configured for {}".format(analysis_id))
         return failed_tests
 
-    for unit_test in analysis["Tests"]:
-        try:
-            entry = unit_test.get("Resource") or unit_test["Log"]
-            log_type = entry.get("p_log_type", "")
-            # set up each test case, including any relevant data models
-            test_case = TestCase(entry, analysis_data_models.get(log_type))
-            result = analysis_funcs["run"](test_case)
-        except KeyError as err:
-            logging.warning("KeyError: {%s}", err)
-            failed_tests[analysis.get("PolicyID") or analysis["RuleID"]].append(unit_test["Name"])
-            continue
-        except Exception as err:  # pylint: disable=broad-except
-            # Catch arbitrary exceptions raised by user code
-            logging.warning("Unexpected exception: {%s}", err)
-            failed_tests[analysis.get("PolicyID") or analysis["RuleID"]].append(unit_test["Name"])
-            continue
-
-        # using a dictionary to map between the tests and their outcomes
-        # assume the test passes (default "PASS")
-        # until failure condition is found (set to "FAIL")
-        test_result: Dict[Any, str] = defaultdict(lambda: "PASS")
-        # check expected result
-        if result != unit_test["ExpectedResult"]:
-            test_result["outcome"] = "FAIL"
-
-        # check dedup and title function return non-None
-        # Only applies to rules which match an incoming event
-        if unit_test["ExpectedResult"]:
-            for func in ["dedup", "title"]:
-                if analysis_funcs.get(func):
-                    if not analysis_funcs[func](test_case):
-                        test_result[func] = "FAIL"
-                        test_result["outcome"] = "FAIL"
-
-        if test_result["outcome"] == "FAIL":
-            failed_tests[analysis.get("PolicyID") or analysis["RuleID"]].append(unit_test["Name"])
-
-        # print results
-        print("\t[{}] {}".format(test_result["outcome"], unit_test["Name"]))
-        for func in ["dedup", "title"]:
-            if analysis_funcs.get(func) and unit_test["ExpectedResult"]:
-                print(
-                    "\t\t[{}] [{}] {}".format(
-                        test_result[func], func, analysis_funcs[func](test_case)
-                    )
-                )
+    failed_tests = _run_tests(analysis, analysis_funcs, analysis_data_models, failed_tests)
 
     if minimum_tests > 1 and not (
         [x for x in analysis["Tests"] if x["ExpectedResult"]]
@@ -922,6 +968,160 @@ def run_tests(
     return failed_tests
 
 
+def _run_tests(
+    analysis: Dict[str, Any],
+    analysis_funcs: Dict[str, Any],
+    analysis_data_models: Dict[str, DataModel],
+    failed_tests: DefaultDict[str, list],
+) -> DefaultDict[str, list]:
+    for unit_test in analysis["Tests"]:
+        try:
+            entry = unit_test.get("Resource") or unit_test["Log"]
+            log_type = entry.get("p_log_type", "")
+            mocks = unit_test.get("Mocks")
+            mock_methods: Dict[str, Any] = {}
+            if mocks:
+                mock_methods = {
+                    each_mock["objectName"]: MagicMock(return_value=each_mock["returnValue"])
+                    for each_mock in mocks
+                    if "objectName" in each_mock and "returnValue" in each_mock
+                }
+            if analysis.get("PolicyID") is not None:
+                # Policies use plain dict objects as input
+                test_case = entry
+            else:
+                # Set up each test case, including any relevant data models
+                test_case = TestCase(entry, analysis_data_models.get(log_type))
+            if mock_methods:
+                with patch.multiple(analysis_funcs["module"], **mock_methods):
+                    result = analysis_funcs["run"](test_case)
+            else:
+                result = analysis_funcs["run"](test_case)
+        except (AttributeError, KeyError) as err:
+            logging.warning("AttributeError: {%s}", err)
+            logging.debug(str(err), exc_info=err)
+            failed_tests[analysis.get("PolicyID") or analysis["RuleID"]].append(unit_test["Name"])
+            continue
+        except Exception as err:  # pylint: disable=broad-except
+            # Catch arbitrary exceptions raised by user code
+            logging.warning("Unexpected exception: {%s}", err)
+            logging.debug(str(err), exc_info=err)
+            failed_tests[analysis.get("PolicyID") or analysis["RuleID"]].append(unit_test["Name"])
+            continue
+
+        # using a dictionary to map between the tests and their outcomes
+        # assume the test passes (default "PASS")
+        # until failure condition is found (set to "FAIL")
+        test_result: Dict[Any, str] = defaultdict(lambda: "PASS")
+        # check expected result
+        if result != unit_test["ExpectedResult"]:
+            test_result["outcome"] = "FAIL"
+            failed_tests[analysis.get("PolicyID") or analysis["RuleID"]].append(unit_test["Name"])
+
+        # print results
+        print("\t[{}] {}".format(test_result["outcome"], unit_test["Name"]))
+
+        # validate reserved function return types and values
+        # Only applies to rules which match an incoming event
+        if unit_test["ExpectedResult"]:
+            verify_result = _verify_test_run(test_case, mock_methods, analysis_funcs)
+            for function_name in verify_result:
+                if verify_result[function_name] != "PASS":
+                    test_result[function_name] = test_result["outcome"] = "FAIL"
+                    failed_tests[analysis.get("PolicyID") or analysis["RuleID"]].append(
+                        f"{unit_test['Name']}:{function_name}"
+                    )
+    return failed_tests
+
+
+def _verify_test_run(
+    test_case: TestCase,
+    mock_methods: Dict[str, MagicMock],
+    analysis_funcs: Dict[str, Any],
+) -> Dict[str, Any]:
+    test_run_result: Dict[str, str] = defaultdict(lambda: "PASS")
+    for func_name, func in analysis_funcs.items():
+        if func_name not in RESERVED_FUNCTIONS:
+            continue
+        try:
+            if mock_methods:
+                with patch.multiple(analysis_funcs["module"], **mock_methods):
+                    func_out = func(test_case)
+                    if not func_out:
+                        test_run_result[func_name] = "FAIL"
+            else:
+                func_out = func(test_case)
+                if not func_out:
+                    test_run_result[func_name] = "FAIL"
+        except Exception as err:  # pylint: disable=broad-except
+            func_out = err
+            test_run_result[func_name] = "FAIL"
+
+        func_out_valid_type = True
+        valid_output, func_out = validate_outputs(func_name, func_out)
+        if not valid_output or isinstance(func_out, (AttributeError, AssertionError)):
+            func_out_valid_type = False
+            test_run_result[func_name] = "FAIL"
+
+        print(
+            f"\t\t[{test_run_result[func_name]}] "
+            f"{'' if func_out_valid_type else '[INVALID TYPE] '}"
+            f"[{func_name}] {func_out}"
+        )
+    return test_run_result
+
+
+def validate_outputs(function_name: str, function_output: Any) -> Tuple[bool, Any]:
+    # Defaults to valid and function output
+    # Invalidating criteria will overwrite these values
+    is_valid = True
+    output = function_output
+    if function_name == "severity":
+        # Type checking for severity
+        if not isinstance(function_output, str):
+            is_valid = False
+        # Value Validation for severity
+        if str(function_output).upper() not in VALID_SEVERITIES:
+            is_valid = False
+            output = AssertionError(
+                f'Expected [{function_name}] to be any of the following: '
+                f'{str(VALID_SEVERITIES)}, got [{str(function_output)}] instead.'
+            )
+    elif function_name == "destinations":
+        # Type checking for destinations
+        if isinstance(function_output, list):
+            # Type checking for elements in list
+            # Note: we cannot perform value validation unless we were to query the outputs-api
+            if not all(isinstance(x, str) for x in function_output):
+                # List contains non-string element, so enumerate the types present for error msg
+                list_types = set()
+                for each_item in function_output:
+                    list_types.add(type(each_item))
+                is_valid = False
+                output = AttributeError(
+                    f'Expected [{function_name}] to return a list of strings, '
+                    f'got types [{list_types}] instead'
+                )
+        else:
+            is_valid = False
+    # For reserved functions besides severity and destinations, we expect a string output
+    else:
+        is_valid = isinstance(function_output, str)
+
+    # Invalid Types, used to generate a corresponding error message
+    if not is_valid and output == function_output:
+        if function_name == "destinations":
+            expected_type = "list"
+        else:
+            expected_type = "string"
+
+        output = AttributeError(
+            f'Expected [{function_name}] to return a [{expected_type}], '
+            f'got [{type(function_output)}] instead'
+        )
+    return is_valid, output
+
+
 def setup_parser() -> argparse.ArgumentParser:
     # pylint: disable=too-many-statements,too-many-locals
     # setup dictionary of named args for some common arguments across commands
@@ -977,7 +1177,7 @@ def setup_parser() -> argparse.ArgumentParser:
         + "managing Panther policies and rules.",
         prog="panther_analysis_tool",
     )
-    parser.add_argument("--version", action="version", version="panther_analysis_tool 0.5.1")
+    parser.add_argument("--version", action="version", version="panther_analysis_tool 0.6.0")
     parser.add_argument("--debug", action="store_true", dest="debug")
     subparsers = parser.add_subparsers()
 
@@ -1025,13 +1225,13 @@ def setup_parser() -> argparse.ArgumentParser:
     )
     publish_parser.add_argument(
         "--github-owner",
-        help="The github owner of the repsitory",
+        help="The github owner of the repository",
         type=str,
         default="panther-labs",
     )
     publish_parser.add_argument(
         "--github-repository",
-        help="The github repsitory name",
+        help="The github repository name",
         type=str,
         default="panther-analysis",
     )
